@@ -7,12 +7,13 @@ code, while avoiding time leaks.
 """
 
 
-__version__ = "0.4.5"
+__version__ = "0.5.0"
 
 from . import backtester
 from . import strategy
 from . import fees
 from . import data_providers
 from . import metrics
+from . import exceptions
 
-__all__ = ["backtester", "strategy", "fees", "data_providers", "metrics"]
+__all__ = ["backtester", "strategy", "fees", "data_providers", "metrics", "exceptions"]
\ No newline at end of file
