@@ -12,6 +12,8 @@ from warnings import warn
 import os
 from IPython.display import clear_output
 from dataclasses import dataclass
+import multiprocessing
+import threading
 from collections.abc import MutableSequence
 
 from .data_providers import (
@@ -21,16 +23,18 @@ from .data_providers import (
     PriceUnavailableError,
     DataProvider,
 )
-from .fees import NoFee, Fee
+from .fees import NoFee, Fee, InsufficientCapitalError
 from .metrics import (
     MaxDrawdown,
     AnnualReturn,
     PortfolioValue,
     DailyProfitLoss,
     TotalValue,
+    TotalReturn,
     Metric,
 )
 from .strategy import Strategy, BuyAndHold
+from .exceptions import BacktestRunError, LongShortLiquidationError, NegativeValueError
 
 try:
     from IPython import display
@@ -75,14 +79,18 @@ class StrategySequence:
 
     def __getitem__(self, index: Union[str, int]):
         if isinstance(index, int):
-            self.bt._get_bts()[index]
+            bt = self.bt._get_bts()[index]
+            bt._from_sequence = True
+            return bt
         elif isinstance(index, str):
             for i, bt in enumerate(self.bt._get_bts()):
                 if bt.name is not None:
                     if bt.name == index:
+                        bt._from_sequence = True
                         return bt
                 else:
                     if f"Backtest {i}" == index:
+                        bt._from_sequence = True
                         return bt
             raise IndexError
 
@@ -103,23 +111,13 @@ def _cls():
     os.system("cls" if os.name == "nt" else "clear")
 
 
-class LongShortLiquidationError(Exception):
-    def __init__(self, message):
-        self.message = message
-
-
-class InsufficientCapitalError(Exception):
-    def __init__(self, message):
-        self.message = message
-
-
 class Position:
     def __init__(self, bt, symbol, date, event, num_shares):
         self.symbol = symbol
         self.date = date
         self.event = event
         self.num_shares_int = num_shares
-        self.init_price = bt.price(symbol)
+        self.start_price = bt.price(symbol)
         self.bt = bt
         self.frozen = False
         self.df_cols = [
@@ -159,31 +157,31 @@ class Position:
     def value(self):
         if self.short:
             old_val = self.initial_value
-            cur_val = self.num_shares * self.bt.price(self.symbol)
+            cur_val = self.num_shares * self.price
             return old_val + (old_val - cur_val)
         if self.long:
-            return self.num_shares * self.bt.price(self.symbol)
+            return self.num_shares * self.price
 
     @property
     def price(self):
         if self.frozen:
-            return self.bt.prices[self.symbol, self.end_date, self.end_event]
+            return self.bt.prices[self.symbol, self.end_date][self.end_event]
         else:
             return self.bt.price(self.symbol)
 
     @property
     def value_pershare(self):
         if self.long:
-            return self.bt.price(self.symbol)
+            return self.price
         if self.short:
-            return self.init_price + (self.init_price - self.bt.price(self.symbol))
+            return self.start_price + (self.start_price - self.price)
 
     @property
     def initial_value(self):
         if self.short:
-            return self.num_shares * self.init_price
+            return self.num_shares * self.start_price
         if self.long:
-            return self.num_shares * self.init_price
+            return self.num_shares * self.start_price
 
     @property
     def profit_loss_pct(self):
@@ -216,6 +214,7 @@ class Position:
         self.frozen = True
         self.end_date = self.bt.current_date
         self.end_event = self.bt.event
+        self.df_cols += ["start_price", "end_date", "end_event"]
 
 
 class Portfolio(MutableSequence):
@@ -252,12 +251,18 @@ class Portfolio(MutableSequence):
             if pos.short:
                 is_short = True
         if is_long and is_short:
+            self.bt._graceful_stop()
             raise LongShortLiquidationError(
                 "liquidating a mix of long and short positions is not possible"
             )
         for pos in copy.copy(self.positions):
             if num_shares == -1 or num_shares >= pos.num_shares:
                 self.bt._available_capital += pos.value
+                if self.bt._available_capital < 0:
+                    self.bt._graceful_stop()
+                    raise NegativeValueError(
+                        f"Tried to liquidate position resulting in negative capital {self.bt._available_capital}."
+                    )
                 self.bt.portfolio._remove(pos)
 
                 pos._freeze()
@@ -351,6 +356,15 @@ class Portfolio(MutableSequence):
                 new_pos.append(pos)
         return Portfolio(self.bt, new_pos)
 
+    def attr(self, attribute):
+        result = [getattr(pos, attribute) for pos in self.positions]
+        if len(result) == 0:
+            return None
+        elif len(result) == 1:
+            return result[0]
+        else:
+            return result
+
 
 class BacktesterBuilder:
     """
@@ -425,7 +439,7 @@ class BacktesterBuilder:
         return self
 
     def metrics(self, metrics: Union[Metric, List[Metric]]) -> "BacktesterBuilder":
-        if type(metrics) == list:
+        if isinstance(metrics, list):
             for m in metrics:
                 for m in metrics:
                     m.bt = self.bt
@@ -458,8 +472,17 @@ class BacktesterBuilder:
         self.bt._live_metrics_every = every
         return self
 
+    def no_live_metrics(self) -> "BacktesterBuilder":
+        scopy = copy.deepcopy(self)
+        scopy.bt._live_metrics = False
+        return scopy
+
     def live_plot(
-        self, every: int = 10, metric: str = "Total Value", event: str = "open",
+        self,
+        every: int = 10,
+        metric: str = "Total Value",
+        figsize: Tuple[float, float] = None,
+        min_y: int = 0
     ) -> "BacktesterBuilder":
         if self.bt._live_metrics:
             self.bt._warn.append(
@@ -472,14 +495,25 @@ class BacktesterBuilder:
         self.bt._live_plot = True
         self.bt._live_plot_every = every
         self.bt._live_plot_metric = metric
-        self.bt._live_plot_event = event
+        self.bt._live_plot_figsize = figsize
+        self.bt._live_plot_min = min_y
         return self
 
+    def no_live_plot(self) -> "BacktesterBuilder":
+        scopy = copy.deepcopy(self)
+        scopy.bt._live_plot = False
+        return scopy
+
     def live_progress(self, every: int = 10) -> "BacktesterBuilder":
         self.bt._live_progress = True
         self.bt._live_progress_every = every
         return self
 
+    def no_live_progress(self) -> "BacktesterBuilder":
+        scopy = copy.deepcopy(self)
+        scopy.bt._live_progress = False
+        return scopy
+
     def compare(
         self,
         strategies: List[
@@ -504,6 +538,10 @@ class BacktesterBuilder:
         self.bt._has_strategies = True
         return self
 
+    def slippage(self, slippage):
+        self.bt._slippage = slippage
+        return self
+
     def build(self) -> "Backtester":
         self.bt._builder = self
         return copy.deepcopy(self.bt)
@@ -511,7 +549,11 @@ class BacktesterBuilder:
 
 class Backtester:
     def __getitem__(self, date_range: slice) -> "Backtester":
-        self = self._builder.build()
+        if self._run_before:
+            raise BacktestRunError(
+                "Backtest has already run, build a new backtester to run again."
+            )
+        self._run_before = True
         if self.assume_nyse:
             self._calendar = "NYSE"
         if date_range.start is not None:
@@ -524,16 +566,26 @@ class Backtester:
             self._warn.append(
                 "backtests with no end date can lead to non-replicable results"
             )
-            end_date = date.today() - relativedelta(days=1)
+            end_date = datetime.date.today() - relativedelta(days=1)
         cal = mcal.get_calendar(self._calendar)
-        if type(start_date) == relativedelta:
-            start_date = date.today() + start_date
-        if type(end_date) == relativedelta:
-            end_date = date.today() + end_date
+        if isinstance(start_date, relativedelta):
+            start_date = datetime.date.today() + start_date
+        if isinstance(end_date, relativedelta):
+            end_date = datetime.date.today() + end_date
         sched = cal.schedule(start_date=start_date, end_date=end_date)
         self._schedule = sched
         self.dates = mcal.date_range(sched, frequency="1D")
+        self.datetimes = []
         self.dates = [d.date() for d in self.dates]
+        for date in self.dates:
+            self.datetimes += [
+                sched.loc[date]["market_open"],
+                sched.loc[date]["market_close"],
+            ]
+
+        if self._has_strategies:
+            self._set_strategies(self._temp_strategies)
+
         return self
 
     def __init__(self):
@@ -553,6 +605,7 @@ class Backtester:
             AnnualReturn(),
             PortfolioValue(),
             TotalValue(),
+            TotalReturn(),
             DailyProfitLoss(),
         ]
         self.metric = {}
@@ -567,6 +620,7 @@ class Backtester:
         self._capital = None
 
         self._live_plot = False
+        self._live_plot_figsize = None
         self._live_metrics = False
         self._live_progress = False
 
@@ -580,12 +634,29 @@ class Backtester:
         self._schedule = None
         self._warn = []
 
-    def _set_self(self):
+        self._add_metrics = {}
+        self._add_metrics_lines = []
+
+        self.datetimes = None
+        self.add_metric_exists = False
+
+        self._run_before = False
+
+        self._last_thread = None
+
+        self._from_sequence = False
+
+        self._slippage = None
+
+    def _set_self(self, new_self=None):
+        if new_self is not None:
+            self = new_self
         self.portfolio.bt = self
         self.trades.bt = self
         self.prices.bt = self
 
         for m in self.metric.values():
+            m.__init__()
             m.bt = self
 
     def _init_iter(self, bt=None):
@@ -605,19 +676,22 @@ class Backtester:
         bt.i = -1
         bt.event = "close"
         if self._live_progress:
-            _live_progress_pbar = tqdm(total=len(self) // 2)
+            _live_progress_pbar = tqdm(total=len(self))
             _cls()
         return self
 
     def _next_iter(self, bt=None):
         if bt is None:
             bt = self
+        if bt.i == len(self):
+            bt._init_iter()
         if bt.event == "open":
             bt.event = "close"
+            bt.i += 1
         elif bt.event == "close":
             try:
                 bt.i += 1
-                bt.current_date = bt.dates[bt.i]
+                bt.current_date = bt.dates[bt.i // 2]
                 bt.event = "open"
             except IndexError:
                 bt.i -= 1
@@ -629,20 +703,30 @@ class Backtester:
                         for metric in strat.metric.values():
                             if metric._single:
                                 metric(write=True)
-                self.plot(self._strategies, last=True)
+                self.plot(self._get_bts(), last=True)
                 for w in self._warn:
                     warn(w)
                 raise StopIteration
         bt._update()
         return bt.current_date, bt.event, bt
 
+    def add_metric(self, key, value):
+        if key not in self._add_metrics:
+            self._add_metrics[key] = (
+                np.repeat(np.nan, len(self)),
+                np.repeat(True, len(self)),
+            )
+        self._add_metrics[key][0][self.i] = value
+        self._add_metrics[key][1][self.i] = False
+
+    def add_line(self, **kwargs):
+        self._add_metrics_lines.append((self.timestamp, kwargs))
+
     @property
     def timestamp(self):
         return self._schedule.loc[self.current_date][f"market_{self.event}"]
 
     def __iter__(self):
-        if self._has_strategies:
-            self._set_strategies(self._temp_strategies)
         return self._init_iter()
 
     def __next__(self):
@@ -693,7 +777,7 @@ class Backtester:
             print()
             print(self._show_live_progress())
 
-    def _show_live_plot(self, bts=None):
+    def _show_live_plot(self, bts=None, start_end=None):
         if not plt_exists:
             self._warn.append(
                 "matplotlib not installed, setting live plotting to false"
@@ -701,27 +785,58 @@ class Backtester:
             self._live_plot = False
             return None
         plot_df = pd.DataFrame()
-        if bts is None:
-            metric = self.metric[self._live_plot_metric].df[self._live_plot_event]
-            plot_df["Backtest"] = metric
+        plot_df["Date"] = self.datetimes
+        plot_df = plot_df.set_index("Date")
+        plot_add_df = plot_df.copy()
+        add_metric_exists = False
+        for i, bt in enumerate(bts):
+            metric = bt.metric[self._live_plot_metric].values
+            name = f"Backtest {i}"
+            if bt.name is not None:
+                name = bt.name
+            plot_df[name] = metric
+            for mkey in bt._add_metrics.keys():
+                add_metric = bt._add_metrics[mkey]
+                plot_add_df[mkey] = np.ma.masked_where(add_metric[1], add_metric[0])
+                if not self.add_metric_exists:
+                    self.add_metric_exists = True
+
+        if self._live_plot_figsize is None:
+            if add_metric_exists:
+                self._live_plot_figsize = (10, 13)
+            else:
+                self._live_plot_figsize = (10, 6.5)
+
+        if self.add_metric_exists:
+            fig, axes = plt.subplots(2, 1, sharex=True, figsize=self._live_plot_figsize)
         else:
-            if not self._no_iter:
-                bts = [self] + bts
-            for i, bt in enumerate(bts):
-                metric = bt.metric[self._live_plot_metric].df[self._live_plot_event]
-                name = f"Backtest {i}"
-                if bt.name is not None:
-                    name = bt.name
-                plot_df[name] = metric
-        fig, ax = plt.subplots()
+            fig, axes = plt.subplots(1, 1, sharex=True, figsize=self._live_plot_figsize)
+            axes = [axes]
+
         if self._live_progress:
-            ax.set_title(str(self._show_live_progress()))
-        plot_df.plot(ax=ax)
+            axes[0].set_title(str(self._show_live_progress()))
+        plot_df.plot(ax=axes[0])
+        if self._live_plot_min is not None:
+            axes[0].set_ylim(bottom=self._live_plot_min)
+
+        if self.add_metric_exists:
+            try:
+                interp_df = plot_add_df.interpolate(method="linear")
+                interp_df.plot(ax=axes[1], cmap="Accent")
+                for line in self._add_metrics_lines:
+                    plt.axvline(line[0], **line[1])
+            except TypeError:
+                pass
+
         fig.autofmt_xdate()
-        plt.xlim([self.dates[0], self.dates[-1]])
+        if start_end is not None:
+            plt.xlim([start_end[0], start_end[1]])
+        else:
+            plt.xlim([self.dates[0], self.dates[-1]])
+
         display.clear_output(wait=True)
-        display.display(pl.gcf())
-        plt.close()
+        plt.draw()
+        plt.pause(0.001)
 
     def _show_live_progress(self):
         _live_progress_pbar.n = self.i + 1
@@ -733,18 +848,28 @@ class Backtester:
                 metric(write=True)
         self._capital = self._available_capital + self.metric["Portfolio Value"][-1]
 
-    def _order(self, symbol, capital, as_percent=False):
+    def _graceful_stop(self):
+        if self._last_thread is not None:
+            self._last_thread.join()
+        self.plot(self._get_bts(), last=True)
+
+    def _order(
+        self, symbol, capital, as_percent=False, as_percent_available=False, shares=None
+    ):
+        self._capital = self._available_capital + self.metric["Portfolio Value"]()
         if capital < 0:
             short = True
             capital = (-1) * capital
         else:
             short = False
-        if not as_percent:
+        if not as_percent and not as_percent_available:
             if capital > self._available_capital:
+                self._graceful_stop()
                 raise InsufficientCapitalError("not enough capital available")
-        else:
-            if capital * self._capital > self._available_capital:
+        elif as_percent:
+            if abs(capital * self._capital) > self._available_capital:
                 if not math.isclose(capital * self._capital, self._available_capital):
+                    self._graceful_stop()
                     raise InsufficientCapitalError(
                         f"""
                         not enough capital available:
@@ -752,10 +877,53 @@ class Backtester:
                         with only {self._available_capital} available
                         """
                     )
+        elif as_percent_available:
+            if abs(capital * self._available_capital) > self._available_capital:
+                if not math.isclose(
+                    capital * self._available_capital, self._available_capital
+                ):
+                    self._graceful_stop()
+                    raise InsufficientCapitalError(
+                        f"""
+                        not enough capital available:
+                        ordered {capital} * {self._available_capital}
+                        with only {self._available_capital} available
+                        """
+                    )
         current_price = self.price(symbol)
+        
+        """
+        if self._slippage is not None:
+            self.prices._leak_allowed = True
+
+            today_df = self.prices[symbol, self.current_date]
+
+            if self.event == 'open':
+                next_price = today_df['close']
+                high_low = [today_df['high'], today_df['low']]
+            elif self.event == 'close':
+                tom_df = self.prices[symbol, self.dates[self.i+1 // 2]]
+                next_price = tom_df['open']
+                high_low = [today_df['high'], today_df['low']]
+            current_price = self._slippage(current_price,next_price,high_low)
+
+            self.prices._leak_allowed = False
+        """
+
         if as_percent:
             capital = capital * self._capital
-        num_shares, total = self._trade_cost(current_price, capital)
+        if as_percent_available:
+            capital = capital * self._available_capital
+        try:
+            if shares is None:
+                num_shares, total = self._trade_cost(current_price, capital)
+            else:
+                num_shares, total = self._trade_cost(
+                    current_price, self._available_capital, num_shares=shares
+                )
+        except Exception as e:
+            self._graceful_stop()
+            raise e
         if short:
             num_shares *= -1
         if num_shares != 0:
@@ -764,6 +932,7 @@ class Backtester:
                 Position(self, symbol, self.current_date, self.event, num_shares)
             )
         else:
+            _cls()
             raise Exception(
                 f"""
                 not enough capital specified to order a single share of {symbol}:
@@ -772,11 +941,31 @@ class Backtester:
                 """
             )
 
-    def order_pct(self, symbol, capital):
-        self._order(symbol, capital, as_percent=True)
+    # def order_pct(self, symbol, capital):
+    #    self._order(symbol, capital, as_percent=True)
 
-    def order_abs(self, symbol, capital):
-        self._order(symbol, capital, as_percent=False)
+    # def order_abs(self, symbol, capital):
+    #    self._order(symbol, capital, as_percent=False)
+
+    def long(self, symbol, **kwargs):
+        if "percent" in kwargs:
+            self._order(symbol, kwargs["percent"], as_percent=True)
+        if "absolute" in kwargs:
+            self._order(symbol, kwargs["absolute"])
+        if "percent_available" in kwargs:
+            self._order(symbol, kwargs["percent_available"], as_percent_available=True)
+        if "nshares" in kwargs:
+            self._order(symbol, 1, shares=kwargs["nshares"])
+
+    def short(self, symbol, **kwargs):
+        if "percent" in kwargs:
+            self._order(symbol, -kwargs["percent"], as_percent=True)
+        if "absolute" in kwargs:
+            self._order(symbol, -kwargs["absolute"])
+        if "percent_available" in kwargs:
+            self._order(symbol, -kwargs["percent_available"], as_percent_available=True)
+        if "nshares" in kwargs:
+            self._order(symbol, -1, shares=kwargs["nshares"])
 
     def price(self, symbol):
         try:
@@ -786,6 +975,7 @@ class Backtester:
                 self.prices.rem_cache(symbol)
                 price = self.prices[symbol, self.current_date][self.event]
             except KeyError:
+                self._graceful_stop()
                 raise PriceUnavailableError(
                     symbol,
                     self.current_date,
@@ -794,6 +984,7 @@ class Backtester:
                     """.strip(),
                 )
         if math.isnan(price) or price is None:
+            self._graceful_stop()
             raise PriceUnavailableError(
                 symbol,
                 self.current_date,
@@ -881,7 +1072,7 @@ class Backtester:
     def _set_strategies(
         self, strategies: List[Callable[["Date", str, "Backtester"], None]]
     ):
-        self._strategies_call = strategies
+        self._strategies_call = copy.deepcopy(strategies)
         for strat in strategies:
             new_bt = copy.deepcopy(self)
             new_bt._set_self()
@@ -895,18 +1086,35 @@ class Backtester:
             self._strategies_call[i](*self._next_iter(bt))
 
     def plot(self, bts, last=False):
-        if self._live_plot and (self.i % self._live_plot_every == 0 or last):
+        try:
+            if self._live_plot and (self.i % self._live_plot_every == 0 or last):
+                if self._last_thread is None or not self._last_thread.is_alive():
+                    thr = threading.Thread(target=self._show_live_plot, args=(bts,))
+                    thr.start()
+                    self._last_thread = thr
+                if last:
+                    self._last_thread.join()
+                    self._show_live_plot(bts)
+            if self._live_metrics and (self.i % self._live_metrics_every == 0 or last):
+                self._show_live_metrics(bts)
+            if not (self._live_metrics or self._live_plot) and (
+                self.i % self._live_progress_every == 0 or last
+            ):
+                _cls()
+                print(self._show_live_progress())
+        except:
+            pass
+
+    def show(self, start=None, end=None):
+        bts = self._get_bts()
+        if self._from_sequence:
+            bts = [self]
+        if start is not None or end is not None:
+            self._show_live_plot(bts, [start, end])
+        else:
             self._show_live_plot(bts)
-        if self._live_metrics and (self.i % self._live_metrics_every == 0 or last):
-            self._show_live_metrics(bts)
-        if not (self._live_metrics or self._live_plot) and (
-            self.i % self._live_progress_every == 0 or last
-        ):
-            _cls()
-            print(self._show_live_progress())
 
     def run(self):
-        self._set_strategies(self._temp_strategies)
         self._init_iter()
         self._no_iter = True
 
